""" Constants and validation for metadata attributes that can be set """


import datetime
from collections import namedtuple  # pylint: disable=syntax-error

from .classes import (
    _AttributeFinderInfo,
    _AttributeFinderInfoColor,
    _AttributeFinderInfoStationaryPad,
    _AttributeList,
    _AttributeTagsList,
    _AttributeOSXPhotosDetectedText,
)
from .constants import *
from .datetime_utils import (
    datetime_has_tz,
    datetime_naive_to_local,
    datetime_remove_tz,
    datetime_tz_to_utc,
)

# Information about metadata attributes that can be set
# Each attribute type needs an Attribute namedtuple entry in ATTRIBUTES dict
# To add new entries, create an Attribute namedtuple and create an entry in
# ATTRIBUTES dict where key is short name for the attribute
# Fields in the Attribute namedtuple are:
# name: short name of the attribute -- will also be used as attribute/property
#   in the OSXMetaData class
# short_constant: the short constant name for the attribute
#   (e.g. kMDItemFinderComment)
# constant: the long name of the constant for the attribute
#   (e.g. com.apple.metadata:kMDItemCreator)
#   See https://developer.apple.com/documentation/coreservices/file_metadata/mditem?language=objc
#   for reference on common metadata attributes
# type_: expected type for the attribute, e.g. if Apple says it's a CFString, it'll be python str
#   CFNumber = python float, etc.
# list: (boolean) True if attribute is a list (e.g. a CFArray)
# as_list: (boolean) True if attribute is a single value but stored in a list
#   Note: the only attribute I've seen this on is com.apple.metadata:kMDItemDownloadedDate
# class_: the attribute class to use, e.g. _AttributeList or str
# append: the attribute can be operated on with append
# update: the attribute can be operated on with update
# help: help text for the attribute (for use in command line interface)
# api_help: help text for use in documentation for the programmer who will use the library
#           if None, will use same text as help
# Note: also add short name to __slots__ in __init__.py OSXMetaData
# Note: also add the constant name (e.g. kMDItemDateAdded) to constants.py


Attribute = namedtuple(
    "Attribute",
    [
        "name",
        "short_constant",
        "constant",
        "type_",
        "list",
        "as_list",
        "class_",
        "append",
        "update",
        "help",
        "api_help",
    ],
)

ATTRIBUTES = {
    "authors": Attribute(
        name="authors",
        short_constant="kMDItemAuthors",
        constant=kMDItemAuthors,
        type_=str,
        list=True,
        as_list=False,
        class_=_AttributeList,
        append=True,
        update=True,
        help="The author, or authors, of the contents of the file.  A list of strings.",
        api_help=None,
    ),
    "comment": Attribute(
        name="comment",
        short_constant="kMDItemComment",
        constant=kMDItemComment,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="A comment related to the file.  This differs from the Finder comment, "
        + "kMDItemFinderComment.  A string.",
        api_help=None,
    ),
    "copyright": Attribute(
        name="copyright",
        short_constant="kMDItemCopyright",
        constant=kMDItemCopyright,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="The copyright owner of the file contents.  A string.",
        api_help=None,
    ),
    "creator": Attribute(
        name="creator",
        short_constant="kMDItemCreator",
        constant=kMDItemCreator,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="Application used to create the document content (for example “Word”, “Pages”, "
        + "and so on).  A string.",
        api_help=None,
    ),
    "description": Attribute(
        name="description",
        short_constant="kMDItemDescription",
        constant=kMDItemDescription,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="A description of the content of the resource.  The description may include an abstract, "
        + "table of contents, reference to a graphical representation of content or a "
        + "free-text account of the content.  A string.",
        api_help=None,
    ),
    "downloadeddate": Attribute(
        name="downloadeddate",
        short_constant="kMDItemDownloadedDate",
        constant=kMDItemDownloadedDate,
        type_=datetime.datetime,
        list=True,
        as_list=False,
        class_=_AttributeList,
        append=False,
        update=False,
        help="The date the item was downloaded.  A date in ISO 8601 format, "
        "time and timezone offset are optional: e.g. "
        + "2020-04-14T12:00:00 (ISO 8601 w/o timezone), "
        + "2020-04-14 (ISO 8601 w/o time and time zone), or "
        + "2020-04-14T12:00:00-07:00 (ISO 8601 with timezone offset). "
        + "Times without timezone offset are assumed to be in local timezone.",
        api_help="The date the item was downloaded.  A datetime.datetime object.  "
        + "If datetime.datetime object lacks tzinfo (i.e. it is timezone naive), it "
        + "will be assumed to be in local timezone.",
    ),
    "findercomment": Attribute(
        name="findercomment",
        short_constant="kMDItemFinderComment",
        constant=kMDItemFinderComment,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="Finder comments for this file.  A string.",
        api_help=None,
    ),
    "headline": Attribute(
        name="headline",
        short_constant="kMDItemHeadline",
        constant=kMDItemHeadline,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="A publishable entry providing a synopsis of the contents of the file.  A string.",
        api_help=None,
    ),
    "keywords": Attribute(
        name="keywords",
        short_constant="kMDItemKeywords",
        constant=kMDItemKeywords,
        type_=str,
        list=True,
        as_list=False,
        class_=_AttributeList,
        append=True,
        update=True,
        help="Keywords associated with this file. For example, “Birthday”, “Important”, etc. "
        + "This differs from Finder tags (_kMDItemUserTags) which are keywords/tags shown "
        + 'in the Finder and searchable in Spotlight using "tag:tag_name".  '
        + "A list of strings.",
        api_help=None,
    ),
    "tags": Attribute(
        name="tags",
        short_constant="_kMDItemUserTags",
        constant=_kMDItemUserTags,
        type_=str,
        list=True,
        as_list=False,
        class_=_AttributeTagsList,
        append=True,
        update=True,
        help='Finder tags; searchable in Spotlight using "tag:tag_name".  '
        + "If you want tags/keywords visible in the Finder, use this instead of kMDItemKeywords.  "
        + "A list of Tag objects.",
        api_help=None,
    ),
    "wherefroms": Attribute(
        name="wherefroms",
        short_constant="kMDItemWhereFroms",
        constant=kMDItemWhereFroms,
        type_=str,
        list=True,
        as_list=False,
        class_=_AttributeList,
        append=True,
        update=True,
        help="Describes where the file was obtained from (e.g. URL downloaded from).  "
        + "A list of strings.",
        api_help=None,
    ),
    "finderinfo": Attribute(
        name="finderinfo",
        short_constant="finderinfo",
        constant=FinderInfo,
        type_=str,
        list=False,
        as_list=False,
        class_=_AttributeFinderInfo,
        append=False,
        update=False,
        help="Info set by the Finder, for example tag color.  Colors can also be set by _kMDItemUserTags.  "
        + f"{FinderInfo} is controlled by the Finder and it's recommended you don't directly access this attribute.  "
        + "If you set or remove a color tag via _kMDItemUserTag, osxmetadata will automatically handle "
        + "processing of FinderInfo color tag.",
        api_help=None,
    ),
    "duedate": Attribute(
        name="duedate",
        short_constant="kMDItemDueDate",
        constant=kMDItemDueDate,
        type_=datetime.datetime,
        list=False,
        as_list=False,
        class_=datetime.datetime,
        append=False,
        update=False,
        help="The date the item is due.  A date in ISO 8601 format, "
        "time and timezone offset are optional: e.g. "
        + "2020-04-14T12:00:00 (ISO 8601 w/o timezone), "
        + "2020-04-14 (ISO 8601 w/o time and time zone), or "
        + "2020-04-14T12:00:00-07:00 (ISO 8601 with timezone offset). "
        + "Times without timezone offset are assumed to be in local timezone.",
        api_help="The date the item is due.  A datetime.datetime object.  "
        + "If datetime.datetime object lacks tzinfo (i.e. it is timezone naive), it "
        + "will be assumed to be in local timezone.",
    ),
    "rating": Attribute(
        name="rating",
        short_constant="kMDItemStarRating",
        constant=kMDItemStarRating,
        type_=int,
        list=False,
        as_list=False,
        class_=int,
        append=False,
        update=False,
        help="User rating of this item. "
        + "For example, the stars rating of an iTunes track. An integer.",
        api_help=None,
    ),
    "participants": Attribute(
        name="participants",
        short_constant="kMDItemParticipants",
        constant=kMDItemParticipants,
        type_=str,
        list=True,
        as_list=False,
        class_=_AttributeList,
        append=True,
        update=True,
        help="The list of people who are visible in an image or movie or written about in a document. "
        + "A list of strings.",
        api_help=None,
    ),
    "projects": Attribute(
        name="projects",
        short_constant="kMDItemProjects",
        constant=kMDItemProjects,
        type_=str,
        list=True,
        as_list=False,
        class_=_AttributeList,
        append=True,
        update=True,
        help="The list of projects that this file is part of. For example, if you were working on a movie all of the files could be marked as belonging to the project “My Movie”. "
        + "A list of strings.",
        api_help=None,
    ),
    "version": Attribute(
        name="version",
        short_constant="kMDItemVersion",
        constant=kMDItemVersion,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="The version number of this file. A string.",
        api_help=None,
    ),
    "stationary": Attribute(
        name="stationary",
        short_constant="kMDItemFSIsStationery",
        constant=kMDItemFSIsStationery,
        type_=bool,
        list=False,
        as_list=False,
        class_=bool,
        append=False,
        update=False,
        help="Boolean indicating if this file is stationery. "
        + "Note: this is not what the Finder uses for Stationary Pad flag.  See also 'stationarypad'.",
        api_help=None,
    ),
    "findercolor": Attribute(
        name="findercolor",
        short_constant="findercolor",
        constant=FinderInfo,
        type_=str,
        list=False,
        as_list=False,
        class_=_AttributeFinderInfoColor,
        append=False,
        update=False,
        help="Color tag set by the Finder.  Colors can also be set by _kMDItemUserTags.  "
        + "This is controlled by the Finder and it's recommended you don't directly access this attribute.  "
        + "If you set or remove a color tag via _kMDItemUserTag, osxmetadata will automatically handle "
        + "processing of FinderInfo color tag.",
        api_help=None,
    ),
    "stationarypad": Attribute(
        name="stationarypad",
        short_constant="stationarypad",
        constant=FinderInfo,
        type_=str,
        list=False,
        as_list=False,
        class_=_AttributeFinderInfoStationaryPad,
        append=False,
        update=False,
        help="Marks the file as stationary (a template that can be re-used). "
        + "Setting this to True is the same as checking the 'Stationary Pad' box in Finder Info.",
        api_help=None,
    ),
    "title": Attribute(
        name="title",
        short_constant="kMDItemTitle",
        constant=kMDItemTitle,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="The title of the file. For example, this could be the title of a document, the name of a song, "
        + "or the subject of an email message. A string.",
        api_help=None,
    ),
    "subject": Attribute(
        name="subject",
        short_constant="kMDItemSubject",
        constant=kMDItemSubject,
        type_=str,
        list=False,
        as_list=False,
        class_=str,
        append=True,
        update=False,
        help="Subject of the this item. A string.",
        api_help=None,
    ),
    "osxphotos_detected_text": Attribute(
        name="osxphotos_detected_text",
        short_constant="OSXPhotosDetectedText",
        constant=OSXPhotosDetectedText,
        type_=list,
        list=True,
        as_list=False,
        class_=_AttributeOSXPhotosDetectedText,
        append=True,
        update=True,
        help="Text detected in a photo; used by osxphotos (https://github.com/RhetTbull/osxphotos).",
        api_help=None,
    ),
}

# used for formatting output of --list
_SHORT_NAME_WIDTH = max(len(x) for x in ATTRIBUTES) + 5
_LONG_NAME_WIDTH = max(len(x.constant) for x in ATTRIBUTES.values()) + 10
_CONSTANT_WIDTH = 21 + 5  # currently longest is kMDItemDownloadedDate


ATTRIBUTE_DISPATCH = {}
for attr in ATTRIBUTES.values():
    try:
        ATTRIBUTE_DISPATCH[attr.constant].append(attr.name)
    except KeyError:
        ATTRIBUTE_DISPATCH[attr.constant] = [attr.name]

# also add entries for attributes by constant and short constant
# do this after computing widths above
_temp_attributes = {}
for attribute in ATTRIBUTES.values():
    if attribute.constant in ATTRIBUTES:
        raise ValueError(f"Duplicate attribute in ATTRIBUTES: {attribute}")
    _temp_attributes[attribute.constant] = attribute
    _temp_attributes[attribute.short_constant] = attribute
if _temp_attributes:
    ATTRIBUTES.update(_temp_attributes)


def validate_attribute_value(attribute, value):
    """validate that value is compatible with attribute.type_
    and convert value to correct type
    returns value as type attribute.type_
    value may be a single value or a list depending on what attribute expects
    if value contains None, returns None
    """

    # check to see if we got None
    try:
        if None in value:
            return None
    except TypeError:
        if value is None:
            return None

    try:
        if isinstance(value, str):
            value = [value]
        else:
            iter(value)
    except TypeError:
        value = [value]

    if not attribute.list and len(value) > 1:
        # got a list but didn't expect one
        raise ValueError(
            f"{attribute.name} expects only one value but list of {len(value)} provided"
        )

    if attribute.class_ == _AttributeOSXPhotosDetectedText:
        return _AttributeOSXPhotosDetectedText.validate_value(value)

    new_values = []
    for val in value:
        new_val = None
        if attribute.type_ == str:
            new_val = str(val)
        elif attribute.type_ == float:
            try:
                new_val = float(val)
            except ValueError:
                # todo: should this really raise ValueError?
                raise TypeError(
                    f"{val} cannot be converted to expected type {attribute.type_}"
                )
        elif attribute.type_ == int:
            try:
                new_val = int(val)
            except ValueError:
                raise TypeError(
                    f"{val} cannot be converted to expected type {attribute.type_}"
                )
        elif attribute.type_ == bool:
            try:
                new_val = bool(val)
            except ValueError:
                raise TypeError(
                    f"{val} cannot be converted to expected type {attribute.type_}"
                )
        elif attribute.type_ == datetime.datetime:
            if isinstance(val, datetime.datetime):
                new_val = val
            elif not val:
                new_val = None
            else:
                try:
                    new_val = datetime.datetime.fromisoformat(val)
                except ValueError:
                    raise TypeError(
                        f"{val} cannot be converted to expected type {attribute.type_}"
                    )
            if isinstance(new_val, datetime.datetime):
                if not datetime_has_tz(new_val):
                    # assume it's in local time, so add local timezone,
                    # convert to UTC, then drop timezone
                    new_val = datetime_naive_to_local(new_val)
                # convert to UTC and remove timezone
                new_val = datetime_tz_to_utc(new_val)
                new_val = datetime_remove_tz(new_val)
        elif attribute.type_ == list:
            new_val = val if isinstance(val, (list, tuple)) else [val]
        else:
            raise TypeError(f"Unknown type: {type(val)}")
        new_values.append(new_val)
    
    if attribute.list:
        return new_values
    else:
        return new_values[0]



""" Python package to read and write various Mac OS X metadata 
    such as tags/keywords and Finder comments from files """

from ._version import __version__
from .attributes import ATTRIBUTES
from .constants import (
    FINDER_COLOR_BLUE,
    FINDER_COLOR_GRAY,
    FINDER_COLOR_GREEN,
    FINDER_COLOR_NONE,
    FINDER_COLOR_ORANGE,
    FINDER_COLOR_PURPLE,
    FINDER_COLOR_RED,
    FINDER_COLOR_YELLOW,
    FinderInfo,
    _COLORIDS,
    _COLORNAMES,
    _FINDER_COMMENT_NAMES,
    _MAX_FINDERCOMMENT,
    _MAX_WHEREFROM,
    _VALID_COLORIDS,
    _kMDItemUserTags,
    kMDItemAuthors,
    kMDItemComment,
    kMDItemCopyright,
    kMDItemCreator,
    kMDItemDescription,
    kMDItemDownloadedDate,
    kMDItemFinderComment,
    kMDItemHeadline,
    kMDItemKeywords,
    kMDItemParticipants,
    kMDItemProjects,
    kMDItemStarRating,
    kMDItemSubject,
    kMDItemTitle,
    kMDItemUserTags,
    kMDItemVersion,
    kMDItemWhereFroms,
)
from .debug import _debug, _set_debug
from .findertags import Tag, get_tag_color_name
from .osxmetadata import OSXMetaData



""" Functions for handling Finder tags and colors """

import logging
import pathlib
import plistlib

from osxmetadata.constants import (
    _COLORIDS,
    _COLORNAMES,
    _COLORNAMES_LOWER,
    _MAX_FINDER_COLOR,
    _MIN_FINDER_COLOR,
    FINDER_COLOR_NONE,
)


def get_tag_color_name(colorid):
    """Return name of the Finder color based on ID"""
    # TODO: need to figure out how to do this in locale/language name
    try:
        colorname = _COLORIDS[colorid]
    except KeyError:
        raise ValueError(f"Invalid colorid: {colorid}")
    return colorname


def tag_factory(tag_str):
    """creates a Tag object
    tag_str: (str) tag value in format: 'name,color'
    where name is the name of the tag and color specifies the color ID
    the comma and color are optional;
    if not provided Tag will use color assigned in Finder or FINDER_COLOR_NONE if no color
    e.g. tag_factory("foo") -> Tag("foo")
         tag_factory("test,6") -> Tag("test", 6)"""

    if not isinstance(tag_str, str):
        raise TypeError(f"tag_str must be str not {type(tag_str)}")

    values = tag_str.split(",")
    name = values[0]
    name = name.lstrip().rstrip()
    if len(values) > 2:
        raise ValueError(f"More than one value found after comma: {tag_str}")
    elif len(values) == 1:
        # got name only, check to see if name is also a color and if so assign it
        # TODO: this might not be right/desired in different locale settings
        # but I'm not sure yet how to get the non-English color names used by Finder
        try:
            colorid = _COLORNAMES_LOWER[name.lower()]
            return Tag(_COLORIDS[colorid], colorid)
        except KeyError:
            return Tag(name)
    elif len(values) == 2:
        # got a color, is it a name or number
        color = values[1].lstrip().rstrip().lower()
        try:
            colorid = _COLORNAMES_LOWER[color]
        except KeyError:
            colorid = int(color)
            if colorid not in range(_MAX_FINDER_COLOR + 1):
                raise ValueError(
                    f"color must be in range 0 to {_MAX_FINDER_COLOR} inclusive: {colorid}"
                )
        return Tag(name, colorid)


def get_finder_tags():
    """parses com.apple.finder.plist to get list of Finder tags and associated colors
    returns: dict in form {tag_name: color_id}"""

    # TODO: Is it possible to get a race condition where Finder is writing at same time
    # this code is trying to open the plist file?  If so, what happens?
    plist_file = pathlib.Path(
        str(pathlib.Path.home()) + "/Library/SyncedPreferences/com.apple.finder.plist"
    )
    tags = {}
    try:
        with open(plist_file, "rb") as fp:
            pl = plistlib.load(fp)
            try:
                finder_tags = pl["values"]["FinderTagDict"]["value"]["FinderTags"]
                for tag in finder_tags:
                    try:
                        name = tag["n"]
                        color = tag["l"]
                    except KeyError:
                        # color will not be present if no color
                        color = 0
                    tags[name] = color

            except Exception as e:
                logging.warning(f"Exception while parsing plist file: {e}")
                raise e

    except Exception as e:
        logging.warning(f"Could not open plist file: {plist_file}, exception: {e}")
        raise e

    return tags


class Tag:
    """Information about Finder tags/keywords and associated color labels"""

    def __init__(self, name, *color):
        """Create a new Tag object
        name: (str) name of the tag
        color: (int) color ID of the tag, optional
        If color name is not provided, com.apple.finder.plist will be parsed to see if
        a color is already associated with 'name' and if so, that color will be assigned
        If a color is provided and the Finder's color is different, a warning will be logged
        but the new color ID will be used"""

        # input validation
        # todo: add length validation on name? Need to determine what max tag length is for Finder
        # and if there are any illegal characters
        if type(name) is not str:
            raise TypeError(f"name must be a str not {type(name)}")
        if len(color):
            if len(color) != 1:
                raise ValueError(
                    f"only 1 argument expected for color but {len(color)} provided"
                )
            if type(color[0]) is not int:
                raise TypeError(f"color must be an int not {type(color[0])}")
            if color[0] not in range(_MAX_FINDER_COLOR + 1):
                raise ValueError(
                    f"color must be in range 0 to {_MAX_FINDER_COLOR} inclusive: {color[0]}"
                )

        self._finder_tags = get_finder_tags()
        self._name = name

        # if no color provided use color from Finder if it exists
        if len(color):
            # color was provided
            if name in self._finder_tags and color[0] != self._finder_tags[name]:
                logging.warning(
                    f"assigning color {color[0]} to tag {name} but "
                    f"color {self._finder_tags[name]} already assigned in Finder"
                )
            self._colorid = color[0]
        elif name in self._finder_tags:
            # use color Finder already using
            self._colorid = self._finder_tags[name]
        else:
            self._colorid = FINDER_COLOR_NONE  # set to no color

    @property
    def name(self):
        return self._name

    @property
    def color(self):
        return self._colorid

    def _format(self):
        """formats a tag for writing to _kMDItemUserTags"""
        return f"{self.name}\n{self.color}"

    def __str__(self):
        return f"{self.name}: {_COLORIDS[self.color]}"

    def __repr__(self):
        return f"Tag('{self.name}', {self.color})"

    def __eq__(self, other):
        return (
            isinstance(other, Tag)
            and self.name == other.name
            and self.color == other.color
        )

    def __lt__(self, other):
        if not isinstance(other, Tag):
            raise TypeError(f"< not supported between type Tag and type {type(other)}")

        return self.name < other.name

    def __gt__(self, other):
        if not isinstance(other, Tag):
            raise TypeError(f"< not supported between type Tag and type {type(other)}")

        return self.name > other.name

    def __len__(self):
        return len(self.name)



""" Classes for metadata attribute types """

import collections.abc
import datetime
import json
import plistlib
from plistlib import FMT_BINARY  # pylint: disable=E0611

import bitstring

from .constants import (
    _COLORIDS,
    _COLORNAMES,
    _COLORNAMES_LOWER,
    _MAX_FINDER_COLOR,
    _MIN_FINDER_COLOR,
    _VALID_COLORIDS,
    FINDER_COLOR_NONE,
    _kCOLOR_OFFSET,
    _kSTATIONARYPAD_OFFSET,
)
from .datetime_utils import (
    datetime_naive_to_utc,
    datetime_remove_tz,
    datetime_utc_to_local,
)
from .debug import _debug
from .findertags import Tag


def value_to_bool(value):
    """Return bool True/False for a given value.
    If value is string and is True or 1, return True
    If value is string and is False or 0, return False
    Otherwise if value is numeric or None, return bool(value)
    """
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        return value.lower() in ("true", "1")
    elif isinstance(value, int) or value is not None:
        return bool(value)
    else:
        return False


class _AttributeList(collections.abc.MutableSequence):
    """represents a multi-valued OSXMetaData attribute list"""

    def __init__(self, attribute, xattr_, osxmetadata_obj):
        """initialize object
        attribute: an OSXMetaData Attributes namedtuple
        xattr_: an instance of xattr.xattr
        osxmetadata_obj: instance of OSXMetaData that created this class instance"""
        self._attribute = attribute
        self._attrs = xattr_
        self._md = osxmetadata_obj

        self._constant = attribute.constant

        self.data = []
        self._values = []
        self._load_data()

    def set_value(self, value):
        self.data = value
        self._write_data()

    def get_value(self):
        self._load_data()
        return self.data

    def _load_data(self):
        self._values = []
        try:
            # load the binary plist value
            self._values = plistlib.loads(self._attrs[self._constant])
            if self._values:
                try:
                    self.data = list(self._values)
                    if self._attribute.type_ == datetime.datetime:
                        # add UTC timezone to datetime values
                        # because that's how MacOS stores them
                        # In the plist associated with xattr, times are stored as:
                        # <date>2020-04-14T14:49:22Z</date>
                        self.data = [datetime_naive_to_utc(dt) for dt in self.data]
                        if not self._md._tz_aware:
                            # want datetimes in naive format
                            self.data = [
                                datetime_remove_tz(d_local)
                                for d_local in [
                                    datetime_utc_to_local(d_utc) for d_utc in self.data
                                ]
                            ]
                except TypeError:
                    self.data = set([self._values])
                    if self._attribute.type_ == datetime.datetime:
                        # add UTC timezone to datetime values
                        self.data = {datetime_naive_to_utc(dt) for dt in self.data}
                        if not self._tz_aware:
                            # want datetimes in naive format
                            self.data = {
                                datetime_remove_tz(d_local)
                                for d_local in {
                                    datetime_utc_to_local(d_utc) for d_utc in self.data
                                }
                            }
            else:
                self.data = []
        except KeyError:
            self.data = []

    def sort(self, key=None, reverse=False):
        self._load_data()
        self.data.sort(key=key, reverse=reverse)
        self._write_data()

    def __delitem__(self, index):
        self._load_data()
        self.data.__delitem__(index)
        self._write_data()

    def __getitem__(self, index):
        self._load_data()
        return self.data.__getitem__(index)

    def __len__(self):
        self._load_data()
        return self.data.__len__()

    def __setitem__(self, index, value):
        self._load_data()
        self.data.__setitem__(index, value)
        self._write_data()
        self._load_data()

    def insert(self, index, value):
        self._load_data()
        self.data.insert(index, value)
        self._write_data()

    def _write_data(self):
        # Overwrites the existing attribute values with the iterable of values provided.
        plist = plistlib.dumps(self.data, fmt=FMT_BINARY)
        self._attrs.set(self._constant, plist)

    def __repr__(self):
        # return f"{type(self).__name__}({super().__repr__()})"
        self._load_data()
        return repr(self.data)

    def __eq__(self, other):
        self._load_data()
        return self.data == other

    def __ne__(self, other):
        return not self.__eq__(other)


class _AttributeFinderInfo:
    """represents info stored in com.apple.FinderInfo"""

    @classmethod
    def _str_to_value_dict(cls, value):
        """Convert str in format key1:value1,key2:value2, etc to a dict"""
        if value:
            value = value.split(",")
            value = [v.split(":") for v in value]
            value = {k: v for k, v in value}
        else:
            value = {}

        if "color" in value:
            try:
                color = int(value["color"])
            except ValueError:
                color = _COLORNAMES_LOWER[value["color"].lower()]
            value["color"] = color

        return value

    def __init__(self, attribute, xattr_, osxmetadata_obj):
        """initialize object
        attribute: an OSXMetaData Attributes namedtuple
        xattr_: an instance of xattr.xattr
        osxmetadata_obj: instance of OSXMetaData that created this class instance"""

        self._attribute = attribute
        self._attrs = xattr_
        self._md = osxmetadata_obj

        # valid keys for the finderinfo dict
        self.keys = ["color", "stationarypad"]

        self._constant = attribute.constant

        self.data = None
        self._load_data()

    def set_value(self, value):
        if not isinstance(value, dict):
            raise TypeError(f"value must be a dict, not {type(value)}")

        if any(key not in self.keys for key in value):
            raise ValueError(f"invalid value dict: value may only contain {self.keys}")

        self.data = value
        self._write_data()

    def get_value(self):
        self._load_data()
        return self.data

    def _load_data(self):
        self._tags = {}

        # check color tag stored in com.apple.FinderInfo
        color = self.get_finderinfo_color()
        stationarypad = self.get_finderinfo_stationarypad()

        self.data = {"color": color, "stationarypad": stationarypad}

    def _write_data(self):
        # Overwrites the existing attribute values with the iterable of values provided.
        # TODO: _write_data can get called multiple times
        # e.g. md.tags += [Tag("foo", 0)] will result in
        # append --> insert --> __setattr__ --> set_attribute --> set_value
        # which results in _write_data being called twice in a row

        if self.data is None:
            # nothing to do
            return

        colorid = self.data.get("color", None)
        stationarypad = self.data.get("stationarypad", None)
        if colorid is not None:
            self.set_finderinfo_color(colorid)
        if stationarypad is not None:
            self.set_finderinfo_stationarypad(stationarypad)

    @property
    def color(self):
        self._load_data()
        return self.data["color"]

    @color.setter
    def color(self, value):
        self.set_finderinfo_color(value)

    @property
    def stationarypad(self):
        self._load_data()
        return self.data["stationarypad"]

    @stationarypad.setter
    def stationarypad(self, value):
        self.set_finderinfo_stationarypad(value)

    def asdict(self):
        return self.data

    def get_finderinfo_color(self):
        """get the tag color of a file set via com.apple.FinderInfo
        returns: color id as int, 0 if no color
                or None if com.apple.FinderInfo not set"""

        try:
            finderbits = self._get_finderinfo_bits()
            bits = finderbits[_kCOLOR_OFFSET : _kCOLOR_OFFSET + 3]
            return bits.uint
        except Exception as e:
            return None

    def set_finderinfo_color(self, colorid):
        """set tag color of filename to colorid
        colorid: ID of tag color in range 0 to 7
        """

        if colorid is None:
            colorid = FINDER_COLOR_NONE

        if not _MIN_FINDER_COLOR <= colorid <= _MAX_FINDER_COLOR:
            raise ValueError(f"colorid out of range {colorid}")

        # color is encoded as 3 binary bits
        bits = bitstring.BitArray(uint=colorid, length=3)

        # set color bits
        finderbits = self._get_finderinfo_bits()
        finderbits.overwrite(bits, _kCOLOR_OFFSET)
        self._set_findinfo_bits(finderbits)

    def get_finderinfo_stationarypad(self):
        """get the Stationary Pad bit from com.apple.FinderInfo
        returns: True if Stationary Pad is set, False if not set"""

        try:
            finderbits = self._get_finderinfo_bits()
            bit = finderbits.bin[_kSTATIONARYPAD_OFFSET]
            return bool(int(bit))
        except Exception as e:
            return False

    def set_finderinfo_stationarypad(self, value):
        """set the Stationary Pad flag of com.apple.FinderInfo"""

        value = 1 if value_to_bool(value) else 0

        # stationary pad is encoded as a single bit
        bit = bitstring.BitArray(uint=value, length=1)

        # set color bits
        finderbits = self._get_finderinfo_bits()
        finderbits.overwrite(bit, _kSTATIONARYPAD_OFFSET)
        self._set_findinfo_bits(finderbits)

    def _get_finderinfo_bits(self) -> bitstring.BitArray:
        """Get FinderInfo bits"""

        try:
            finderinfo = self._attrs.get("com.apple.FinderInfo")
            finderbits = bitstring.BitArray(finderinfo)
        except Exception:
            finderbits = bitstring.BitArray(uint=0, length=256)
        return finderbits

    def _set_findinfo_bits(self, bits: bitstring.BitArray):
        """Set FinderInfo bits"""
        self._attrs.set("com.apple.FinderInfo", bits.bytes)

    def __repr__(self):
        self._load_data()
        return repr(self.data)

    def __str__(self):
        self._load_data()
        return f"'{self.data}'"

    def __eq__(self, other):
        self._load_data()
        return self.data == other

    def __getitem__(self, key):
        self._load_data()
        return self.data[key]


class _AttributeFinderInfoColor(_AttributeFinderInfo):
    def __init__(self, attribute, xattr_, osxmetadata_obj):
        super().__init__(attribute, xattr_, osxmetadata_obj)

    def set_value(self, value):
        if not isinstance(value, int):
            raise TypeError(f"value must be a int, not {type(value)}")

        self.data = {"color": value}
        self._write_data()

    def get_value(self):
        self._load_data()
        return self.data.get("color", None)

    def __repr__(self):
        return repr(self.get_value())

    def __eq__(self, other):
        self._load_data()
        return self.data.get("color") == other

    def __ne__(self, other):
        return not self.__eq__(other)


class _AttributeFinderInfoStationaryPad(_AttributeFinderInfo):
    def __init__(self, attribute, xattr_, osxmetadata_obj):
        super().__init__(attribute, xattr_, osxmetadata_obj)

    def set_value(self, value):
        self.data = {"stationarypad": value_to_bool(value)}
        self._write_data()

    def get_value(self):
        self._load_data()
        return self.data.get("stationarypad", None)

    def __repr__(self):
        return repr(self.get_value())

    def __eq__(self, other):
        self._load_data()
        return self.data.get("stationarypad") == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __bool__(self):
        return bool(self.data.get("stationarypad", None))


class _AttributeTagsList(_AttributeList, _AttributeFinderInfo):
    """represents a _kMDItemUserTag attribute list"""

    def set_value(self, value):
        if isinstance(value, list):
            # a list of tag objects
            for val in value:
                if not isinstance(val, Tag):
                    raise TypeError(f"values must be type Tag, not {type(val)}")
            self.data = value
        elif isinstance(value, _AttributeTagsList):
            # another _AttributeTagsList object
            self.data = value.data
        else:
            raise TypeError(f"value must be list of Tag objects not {type(value)}")

        self._write_data()

    def _tag_split(self, tag):
        # Extracts the color information from a Finder tag.

        parts = tag.rsplit("\n", 1)
        if len(parts) == 1:
            return parts[0], 0
        elif (
            len(parts[1]) != 1 or parts[1] not in _VALID_COLORIDS
        ):  # Not a color number
            return tag, 0
        else:
            return parts[0], int(parts[1])

    def _load_data(self):
        self._tags = {}
        try:
            tag_constant = self._attrs[self._constant]
            # load the binary plist value
            self._tagvalues = plistlib.loads(tag_constant)
            for x in self._tagvalues:
                (tag, color) = self._tag_split(x)
                self._tags[tag] = color
        except KeyError:
            self._tags = None
        if self._tags:
            self.data = [Tag(name, color) for name, color in self._tags.items()]
        else:
            self.data = []

        # check color tag stored in com.apple.FinderInfo
        color = self.get_finderinfo_color()
        if color and (self._tags is None or color not in self._tags.values()):
            # have a FinderInfo color that's not in _kMDItemUserTag
            self.data.append(Tag(_COLORIDS[color], color))

    def _write_data(self):
        # Overwrites the existing attribute values with the iterable of values provided.
        # TODO: _write_data can get called multiple times
        # e.g. md.tags += [Tag("foo", 0)] will result in
        # append --> insert --> __setattr__ --> set_attribute --> set_value
        # which results in _write_data being called twice in a row
        self._tagvalues = [tag._format() for tag in self.data]
        plist = plistlib.dumps(self._tagvalues, fmt=FMT_BINARY)
        self._attrs.set(self._constant, plist)

        # also write FinderInfo if required
        # if findercolor in tag set being written, do nothing
        # if findercolor not in tag set being written, overwrite findercolor with first color tag
        # the behavior of Finder is to set findercolor to the most recently set color but we won't necessarily know that
        finder_color = self.get_finderinfo_color()
        tag_colors = [tag.color for tag in self.data]
        if finder_color == FINDER_COLOR_NONE or finder_color not in tag_colors:
            # overwrite FinderInfo color with new color
            # get first non-zero color in tag if there is one
            try:
                color = tag_colors[
                    tag_colors.index(
                        next(filter(lambda x: x != FINDER_COLOR_NONE, tag_colors))
                    )
                ]
            except StopIteration:
                color = FINDER_COLOR_NONE
            self.set_finderinfo_color(color)


class _AttributeJSONList(_AttributeList):
    """represents a list item that stores its value as JSON"""

    def __init__(self, attribute, xattr_, osxmetadata_obj):
        """initialize object
        attribute: an OSXMetaData Attributes namedtuple
        xattr_: an instance of xattr.xattr
        osxmetadata_obj: instance of OSXMetaData that created this class instance"""
        self._attribute = attribute
        self._attrs = xattr_
        self._md = osxmetadata_obj

        self._constant = attribute.constant

        self.data = None
        self._json = None
        self._load_data()

    def set_value(self, value):
        if not isinstance(value, list):
            raise TypeError(f"value must be a list, not {type(value)}")
        self.data = value
        self._write_data()

    def get_value(self):
        self._load_data()
        return self.data

    def _load_data(self):
        try:
            self._json = self._attrs[self._constant]
            self.data = json.loads(self._json)
        except KeyError:
            self.data = None

    def _write_data(self):
        self._json = json.dumps(self.data).encode("utf-8")
        self._attrs.set(self._constant, self._json)


class _AttributeOSXPhotosDetectedText(_AttributeJSONList):
    """represents an OSXPhotos.metadata:detected_text attribute"""

    @classmethod
    def validate_value(cls, value):
        """verify value is valid"""

        if not isinstance(value, (list, str)):
            raise TypeError(f"value must be a list or JSON str, not {type(value)}")

        if value and isinstance(value, str):
            # assume it's JSON
            value = json.loads(value[0])
        elif value:
            if not isinstance(value[0], list):
                value = [value]

            for item in value:
                if not isinstance(item, (list, tuple)):
                    raise TypeError(f"items must be lists or tuples, not {type(item)}")
                if len(item) != 2:
                    raise TypeError(f"items must be of length 2, not {len(item)}")
                if type(item[0]) != str:
                    raise TypeError(f"first item must be a string, not {type(item[0])}")
                if not isinstance(item[1], (int, float)):
                    raise TypeError(
                        f"second item must be a number, not {type(item[1])}"
                    )
                item[1] = float(item[1])
        return value

    def set_value(self, value):
        value = self.validate_value(value)
        return super().set_value(value)

    def __len__(self):
        self._load_data()
        return self.data.__len__() if self.data else 0


import logging

_DEBUG = False

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(filename)s - %(lineno)d - %(message)s",
)

if not _DEBUG:
    logging.disable(logging.DEBUG)


def _get_logger():
    """Used only for testing
    
    Returns:
        logging.Logger object -- logging.Logger object for osxmetadata
    """
    return logging.Logger(__name__)


def _set_debug(debug):
    """ Enable or disable debug logging """
    global _DEBUG
    _DEBUG = debug
    if debug:
        logging.disable(logging.NOTSET)
    else:
        logging.disable(logging.DEBUG)


def _debug():
    """ returns True if debugging turned on (via _set_debug), otherwise, false """
    return _DEBUG



""" OSXMetaData class to read and write various Mac OS X metadata 
    such as tags/keywords and Finder comments from files """


import base64
import datetime
import json
import logging
import os.path
import pathlib
import plistlib

# plistlib creates constants at runtime which causes pylint to complain
from plistlib import FMT_BINARY  # pylint: disable=E0611

import applescript
import xattr

from ._version import __version__
from .attributes import ATTRIBUTES, Attribute, validate_attribute_value
from .classes import _AttributeOSXPhotosDetectedText, _AttributeList, _AttributeTagsList
from .constants import (
    _FINDER_COMMENT_NAMES,
    FINDER_COLOR_NONE,
    FinderInfo,
    _kMDItemUserTags,
)
from .datetime_utils import (
    datetime_naive_to_utc,
    datetime_remove_tz,
    datetime_utc_to_local,
)
from .debug import _debug, _get_logger, _set_debug
from .findertags import Tag, get_tag_color_name

# TODO: What to do about colors
# TODO: check what happens if OSXMetaData.__init__ called with invalid file--should result in error but saw one case where it didn't
# TODO: cleartags does not always clear colors--this is a new behavior, did Mac OS change something in implementation of colors?

# all attributes that are part of com.apple.FinderInfo
_FINDERINFO_ATTRIBUTES = ["finderinfo", "findercolor", "stationarypad"]

# all attributes that use an Attribute class (defined in attributes.py)
_ATTRIBUTE_CLASS_ATTRIBUTES = [
    "tags",
    "osxphotos_detected_text",
    *_FINDERINFO_ATTRIBUTES,
]

# all attributes related to stationary pad
_FINDERINFO_STATIONARYPAD_ATTRIBUTES = ["finderinfo", "stationarypad"]

# all attributes related to Finder color
_FINDERINFO_COLOR_ATTRIBUTES = ["finderinfo", "findercolor"]

# all attributes that are included in finderinfo
_FINDERINFO_SUB_ATTRIBUTES = ["findercolor", "stationarypad"]

# AppleScript for manipulating Finder comments
_scpt_set_finder_comment = applescript.AppleScript(
    """
            on run {path, fc}
	            set thePath to path
	            set theComment to fc
	            tell application "Finder" to set comment of (POSIX file thePath as alias) to theComment
            end run
            """
)

_scpt_clear_finder_comment = applescript.AppleScript(
    """
            on run {path}
	            set thePath to path
	            set theComment to missing value
	            tell application "Finder" to set comment of (POSIX file thePath as alias) to theComment
            end run
            """
)


class OSXMetaData:
    """Create an OSXMetaData object to access file metadata"""

    __slots__ = [
        "__init",
        "_attrs",
        "_fname",
        "_posix_name",
        "_tz_aware",
        "authors",
        "comment",
        "copyright",
        "creator",
        "description",
        "downloadeddate",
        "duedate",
        "findercolor",
        "findercomment",
        "finderinfo",
        "headline",
        "keywords",
        "participants",
        "projects",
        "rating",
        "stationary",
        "stationarypad",
        "tags",
        "version",
        "wherefroms",
        "osxphotos_detected_text",
    ]

    def __init__(self, fname, tz_aware=False):
        """Create an OSXMetaData object to access file metadata
        fname: filename to operate on
        timezone_aware: bool; if True, date/time attributes will return
                  timezone aware datetime.datetime attributes; if False (default)
                  date/time attributes will return timezone naive objects"""
        self._fname = pathlib.Path(fname)
        self._posix_name = self._fname.resolve().as_posix()
        self._tz_aware = tz_aware

        if not self._fname.exists():
            raise FileNotFoundError("file does not exist: ", fname)

        self._attrs = xattr.xattr(self._fname)

        # create property classes for the multi-valued attributes
        # tags get special handling due to color labels
        # ATTRIBUTES contains both long and short names, want only the short names (attribute.name)
        for name in {attribute.name for attribute in ATTRIBUTES.values()}:
            attribute = ATTRIBUTES[name]
            if attribute.class_ not in [str, float, int, bool, datetime.datetime]:
                super().__setattr__(
                    name, attribute.class_(attribute, self._attrs, self)
                )

        # Done with initialization
        self.__init = True

    @property
    def name(self):
        """POSIX path of the file OSXMetaData is operating on"""
        return self._fname.resolve().as_posix()

    @property
    def tz_aware(self):
        """returns the timezone aware flag"""
        return self._tz_aware

    @tz_aware.setter
    def tz_aware(self, tz_flag):
        """sets the timezone aware flag
        tz_flag: bool"""
        self._tz_aware = tz_flag

    def asdict(self, all_=False, encode=True):
        """Return dict with all attributes for this file

        Args:
            all_: bool, if True, returns all attributes including those that osxmetadata knows nothing about
            encode: bool, if True, encodes values for unknown attributes with base64, otherwise leaves the values as raw bytes

        Returns:
            dict with attributes for this file
        """

        attribute_list = self._list_attributes() if all_ else self.list_metadata()
        dict_data = {
            "_version": __version__,
            "_filepath": self._posix_name,
            "_filename": self._fname.name,
        }

        for attr in attribute_list:
            try:
                attribute = ATTRIBUTES[attr]
                if attribute.name == "tags":
                    tags = self.get_attribute(attribute.name)
                    value = [[tag.name, tag.color] for tag in tags]
                    dict_data[attribute.constant] = value
                elif attribute.constant == FinderInfo:
                    value = self.finderinfo.asdict()
                    dict_data[attribute.constant] = value
                elif attribute.type_ == datetime.datetime:
                    # need to convert datetime.datetime to string to serialize
                    value = self.get_attribute(attribute.name)
                    if type(value) == list:
                        value = [v.isoformat() for v in value]
                    else:
                        value = value.isoformat()
                    dict_data[attribute.constant] = value
                else:
                    # get raw value
                    dict_data[attribute.constant] = self.get_attribute(attribute.name)
            except KeyError:
                # an attribute osxmetadata doesn't know about
                if all_:
                    try:
                        value = self._attrs[attr]
                        # convert value to base64 encoded ascii
                        if encode:
                            value = base64.b64encode(value).decode("ascii")
                        dict_data[attr] = value
                    except KeyError as e:
                        # value disappeared between call to _list_attributes and now
                        pass
        return dict_data

    def to_json(self, all_=False):
        """Returns a string in JSON format for all attributes in this file

        Args:
            all_: bool; if True, also restores attributes not known to osxmetadata (generated with asdict(all_=True, encode=True) )
        """
        dict_data = self.asdict(all_=all_)
        return json.dumps(dict_data)

    def _restore_attributes(self, attr_dict, all_=False):
        """restore attributes from attr_dict
        for each attribute in attr_dict, will set the attribute
        will not clear/erase any attributes on file that are not in attr_dict

        Args:
            attr_dict: an attribute dict as produced by OSXMetaData.asdict()
            all_: bool; if True, also restores attributes not known to osxmetadata (generated with asdict(all_=True, encode=True) )
        """

        for key, val in attr_dict.items():
            if key.startswith("_"):
                # skip private keys like _version and _filepath
                continue
            try:
                if key == _kMDItemUserTags:
                    if not isinstance(val, list):
                        raise TypeError(
                            f"expected list for attribute {key} but got {type(val)}"
                        )
                    self.set_attribute(key, [Tag(*tag_val) for tag_val in val])
                elif key == FinderInfo:
                    if not isinstance(val, dict):
                        raise TypeError(
                            f"expected dict for attribute {key} but got {type(val)}"
                        )
                    self.set_attribute(key, val)
                elif key in ATTRIBUTES:
                    self.set_attribute(key, val)
                elif all_:
                    self._attrs.set(key, base64.b64decode(val))
            except Exception as e:
                logging.warning(
                    f"Unable to restore attribute {key} for {self._fname}: {e}"
                )

    def get_attribute(self, attribute_name):
        """load attribute and return value or None if attribute was not set
        attribute_name: name of attribute
        """

        attribute = ATTRIBUTES[attribute_name]

        # user tags and finderinfo need special processing
        if attribute.name in _ATTRIBUTE_CLASS_ATTRIBUTES:
            return getattr(self, attribute.name).get_value()

        # must be a "normal" metadata attribute
        try:
            plist = plistlib.loads(self._attrs[attribute.constant])
        except KeyError:
            plist = None

        # add UTC to any datetime.datetime objects because that's how MacOS stores them
        # In the plist associated with extended metadata attributes, times are stored as:
        # <date>2020-04-14T14:49:22Z</date>
        if plist and isinstance(plist, list):
            if isinstance(plist[0], datetime.datetime):
                plist = [datetime_naive_to_utc(d) for d in plist]
                if not self._tz_aware:
                    # want datetimes in naive format
                    plist = [
                        datetime_remove_tz(d_local)
                        for d_local in [datetime_utc_to_local(d_utc) for d_utc in plist]
                    ]
        elif isinstance(plist, datetime.datetime):
            plist = datetime_naive_to_utc(plist)
            if not self._tz_aware:
                # want datetimes in naive format
                plist = datetime_remove_tz(datetime_utc_to_local(plist))

        if attribute.as_list and isinstance(plist, list):
            return plist[0]
        else:
            return plist

    def get_attribute_str(self, attribute_name):
        """returns a string representation of attribute value
        e.g. if attribute is a datedate.datetime object, will
        format using datetime.isoformat()
        attribute_name: name of attribute"""
        
        attribute = ATTRIBUTES[attribute_name]
        value = self.get_attribute(attribute_name)
        if attribute.class_ == _AttributeOSXPhotosDetectedText:
            return json.dumps(value)
        
        if type(value) in [list, set]:
            if value:
                if type(value[0]) == datetime.datetime:
                    new_value = [v.isoformat() for v in value]
                    return str(new_value)
                elif isinstance(value[0], Tag):
                    new_value = [
                        f"{tag.name},{get_tag_color_name(tag.color)}"
                        if tag.color
                        else f"{tag.name}"
                        for tag in value
                    ]
                    return str(new_value)
            return [str(val) for val in value]
        else:
            if type(value) == datetime.datetime:
                return value.isoformat()
            elif isinstance(value, Tag):
                return (
                    f"{value.name},{get_tag_color_name(value.color)}"
                    if value.color
                    else f"{value.name}"
                )
            return str(value)

    def set_attribute(self, attribute_name, value):
        """write attribute to file with value
        attribute_name: an osxmetadata Attribute name
        value: value to store in attribute"""
        attribute = ATTRIBUTES[attribute_name]

        # user tags and Finder info need special processing
        if attribute.name in _ATTRIBUTE_CLASS_ATTRIBUTES:
            return getattr(self, attribute.name).set_value(value)

        # verify type is correct
        value = validate_attribute_value(attribute, value)

        if attribute.name in _FINDER_COMMENT_NAMES:
            # Finder Comment needs special handling
            # code following will also set the attribute for Finder Comment
            self.set_finder_comment(self._posix_name, value)
        elif attribute.class_ in [
            _AttributeList,
            _AttributeTagsList,
        ]:
            getattr(self, attribute.name).set_value(value)
        else:
            # must be a normal scalar (e.g. str, float)
            plist = plistlib.dumps(value, fmt=FMT_BINARY)
            self._attrs.set(attribute.constant, plist)

    def update_attribute(self, attribute_name, value):
        """Update attribute with union of itself and value
        (this avoids adding duplicate values to attribute)
        attribute: an osxmetadata Attribute name
        value: value to append to attribute"""
        return self.append_attribute(attribute_name, value, update=True)

    def append_attribute(self, attribute_name, value, update=False):
        """append value to attribute
        attribute_name: an osxmetadata Attribute name
        value: value to append to attribute
        update: (bool) if True, update instead of append (e.g. avoid adding duplicates)
                (default is False)"""

        attribute = ATTRIBUTES[attribute_name]

        # start with existing values
        new_value = self.get_attribute(attribute.name)

        # user tags need special processing to normalize names
        if attribute.name == "tags":
            if not isinstance(new_value, list) or not isinstance(value, list):
                raise TypeError(
                    f"tags expects values in list: {type(new_value)}, {type(value)}"
                )

            if update:
                # verify not already in values
                for val in value:
                    if val not in new_value:
                        new_value.append(val)
            else:
                new_value.extend(value)
            return self.tags.set_value(new_value)
        if attribute.name in _FINDERINFO_ATTRIBUTES:
            raise ValueError(f"cannot append or update {attribute.name}")

        value = validate_attribute_value(attribute, value)

        if (
            attribute.class_ == _AttributeOSXPhotosDetectedText
            and new_value is None
            or attribute.class_ != _AttributeOSXPhotosDetectedText
            and attribute.list
            and new_value is None
        ):
            new_value = value
        elif attribute.list:
            new_value = list(new_value)
            if update:
                for val in value:
                    if val not in new_value:
                        new_value.append(val)
            elif attribute.class_ == _AttributeOSXPhotosDetectedText:
                for val in value:
                    new_value.append(val)
            else:
                new_value.extend(value)
        else:
            # scalar value
            if update:
                raise AttributeError(f"Cannot use update on {attribute.type_}")
            if new_value is None:
                new_value = value
            else:
                new_value += value

        try:
            if attribute.name in _FINDER_COMMENT_NAMES:
                # Finder Comment needs special handling
                self.set_finder_comment(self._posix_name, new_value)
            elif attribute.class_ in [
                _AttributeList,
                _AttributeTagsList,
                _AttributeOSXPhotosDetectedText,
            ]:
                # if tags, set_value will normalize
                getattr(self, attribute.name).set_value(new_value)
            else:
                plist = plistlib.dumps(new_value, fmt=FMT_BINARY)
                self._attrs.set(attribute.constant, plist)
        except Exception as e:
            # todo: should catch this or not?
            raise e

        return new_value

    def remove_attribute(self, attribute_name, value):
        """remove a value from attribute, raise ValueError if attribute does not contain value
        only applies to multi-valued attributes, otherwise raises TypeError
        attribute_name: name of OSXMetaData attribute"""

        attribute = ATTRIBUTES[attribute_name]

        if not attribute.list:
            raise TypeError("remove only applies to multi-valued attributes")

        if attribute.name == "tags":
            # tags need special processing
            self.tags.remove(value)
        else:
            values = self.get_attribute(attribute.name)
            values = list(values)
            values.remove(value)
            self.set_attribute(attribute.name, values)

    def discard_attribute(self, attribute_name, value):
        """remove a value from attribute, unlike remove, does not raise exception
        if attribute does not contain value
        only applies to multi-valued attributes, otherwise raises TypeError
        attribute_name: name of OSXMetaData attribute"""

        attribute = ATTRIBUTES[attribute_name]

        if not attribute.list:
            raise TypeError("discard only applies to multi-valued attributes")

        values = self.get_attribute(attribute.name)
        try:
            values.remove(value)
            self.set_attribute(attribute.name, values)
        except Exception:
            pass

    def clear_attribute(self, attribute_name):
        """clear anttribute (remove it from the file)
        attribute_name: name of OSXMetaData attribute"""

        attribute = ATTRIBUTES[attribute_name]

        try:
            if attribute.name in _FINDER_COMMENT_NAMES:
                # Finder Comment needs special handling
                # code following will also clear the attribute for Finder Comment
                self.clear_finder_comment(self._posix_name)

            if attribute.name in ["tags", *_FINDERINFO_COLOR_ATTRIBUTES]:
                # don't clear the entire FinderInfo attribute, just delete the bits we know about
                self.finderinfo.set_finderinfo_color(FINDER_COLOR_NONE)

            if attribute.name in _FINDERINFO_STATIONARYPAD_ATTRIBUTES:
                self.finderinfo.set_finderinfo_stationarypad(False)

            if attribute.name not in _FINDERINFO_ATTRIBUTES:
                # remove the entire attribute
                self._attrs.remove(attribute.constant)
        except (IOError, OSError):
            # TODO: fix this try/except handling
            pass

    def _list_attributes(self):
        """list all the attributes set on the file"""
        return self._attrs.list()

    def list_metadata(self):
        """list the Apple metadata attributes set on the file:
        e.g. those in com.apple.metadata namespace"""
        # also lists com.osxmetadata.test used for debugging
        mdlist = self._list_attributes()
        mdlist = [
            md
            for md in mdlist
            if md.startswith("com.apple.metadata")
            or md.startswith("com.apple.FinderInfo")
            or md.startswith("com.osxmetadata.test")
            or md.startswith("osxphotos.metadata")
        ]
        return mdlist

    def set_finder_comment(self, path, comment):
        """set finder comment for object path (file or directory)
        path: path to file or directory in posix format
        comment: comment string
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"Could not find {path}")

        if comment:
            _scpt_set_finder_comment.run(path, comment)
            plist = plistlib.dumps(comment, fmt=FMT_BINARY)
            self._attrs.set(ATTRIBUTES["findercomment"].constant, plist)
        else:
            self.clear_finder_comment(path)

    def clear_finder_comment(self, path):
        """clear finder comment for object path (file or directory)
        path: path to file or directory in posix format
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"Could not find {path}")

        _scpt_clear_finder_comment.run(path)
        try:
            self._attrs.remove(ATTRIBUTES["findercomment"].constant)
        except (IOError, OSError):
            # exception raised if attribute not found and attempt to remove it
            pass

    def __getattr__(self, name):
        """if attribute name is in ATTRIBUTE dict, return the value"""
        if name in ATTRIBUTES:
            return self.get_attribute(name)
        raise AttributeError(f"{name} is not an attribute")

    def __setattr__(self, name, value):
        """if object is initialized and name is an attribute in ATTRIBUTES,
        set the attribute to value
        if value value is None, will clear (delete) the attribute and all associated values
        if name is not a metadata attribute, assume it's a normal class attribute and pass to
        super() to handle"""
        try:
            if self.__init:
                # already initialized
                attribute = ATTRIBUTES[name]
                if value is None:
                    self.clear_attribute(attribute.name)
                else:
                    self.set_attribute(attribute.name, value)
        except (KeyError, AttributeError):
            super().__setattr__(name, value)



""" Constants and definitions used by osxmetadata """

# color labels
_COLORNAMES = {
    "None": 0,
    "Gray": 1,
    "Green": 2,
    "Purple": 3,
    "Blue": 4,
    "Yellow": 5,
    "Red": 6,
    "Orange": 7,
}

_COLORNAMES_LOWER = {
    "none": 0,
    "gray": 1,
    "green": 2,
    "purple": 3,
    "blue": 4,
    "yellow": 5,
    "red": 6,
    "orange": 7,
}

_COLORIDS = {
    0: "None",
    1: "Gray",
    2: "Green",
    3: "Purple",
    4: "Blue",
    5: "Yellow",
    6: "Red",
    7: "Orange",
}

FINDER_COLOR_NONE = 0
FINDER_COLOR_GRAY = 1
FINDER_COLOR_GREEN = 2
FINDER_COLOR_PURPLE = 3
FINDER_COLOR_BLUE = 4
FINDER_COLOR_YELLOW = 5
FINDER_COLOR_RED = 6
FINDER_COLOR_ORANGE = 7

_MIN_FINDER_COLOR = 0
_MAX_FINDER_COLOR = 7

_VALID_COLORIDS = "01234567"
_MAX_FINDERCOMMENT = 750  # determined through trial & error with Finder
_MAX_WHEREFROM = (
    1024  # just picked something....todo: need to figure out what max length is
)

# offset of tag color in com.apple.FinderInfo xattr
# tag color is 3 bits
_kCOLOR_OFFSET = 76

# offset of stationary pad bit in com.apple.FinderInfo xattr
_kSTATIONARYPAD_OFFSET = 68

# _TAGS = "com.apple.metadata:_kMDItemUserTags"
# _FINDER_COMMENT = "com.apple.metadata:kMDItemFinderComment"
# _WHERE_FROM = "com.apple.metadata:kMDItemWhereFroms"
# _DOWNLOAD_DATE = "com.apple.metadata:kMDItemDownloadedDate"


FinderInfo = "com.apple.FinderInfo"
_kMDItemUserTags = "com.apple.metadata:_kMDItemUserTags"
kMDItemAuthors = "com.apple.metadata:kMDItemAuthors"
kMDItemComment = "com.apple.metadata:kMDItemComment"
kMDItemCopyright = "com.apple.metadata:kMDItemCopyright"
kMDItemCreator = "com.apple.metadata:kMDItemCreator"
kMDItemDescription = "com.apple.metadata:kMDItemDescription"
kMDItemDownloadedDate = "com.apple.metadata:kMDItemDownloadedDate"
kMDItemDueDate = "com.apple.metadata:kMDItemDueDate"
kMDItemFSIsStationery = "com.apple.metadata:kMDItemFSIsStationery"
kMDItemFinderComment = "com.apple.metadata:kMDItemFinderComment"
kMDItemHeadline = "com.apple.metadata:kMDItemHeadline"
kMDItemKeywords = "com.apple.metadata:kMDItemKeywords"
kMDItemParticipants = "com.apple.metadata:kMDItemParticipants"
kMDItemProjects = "com.apple.metadata:kMDItemProjects"
kMDItemStarRating = "com.apple.metadata:kMDItemStarRating"
kMDItemSubject = "com.apple.metadata:kMDItemSubject"
kMDItemTitle = "com.apple.metadata:kMDItemTitle"
kMDItemUserTags = "com.apple.metadata:_kMDItemUserTags"
kMDItemVersion = "com.apple.metadata:kMDItemVersion"
kMDItemWhereFroms = "com.apple.metadata:kMDItemWhereFroms"
OSXPhotosDetectedText = "osxphotos.metadata:detected_text"

# Special handling for Finder comments
_FINDER_COMMENT_NAMES = [
    "findercomment",
    "kMDItemFinderComment",
    "com.apple.metadata:kMDItemFinderComment",
]
_TAGS_NAMES = ["tags", "_kMDItemUserTags", "com.apple.metadata:_kMDItemUserTags"]
_FINDERINFO_NAMES = ["finderinfo", "com.apple.FinderInfo"]

_BACKUP_FILENAME = ".osxmetadata.json"

__all__ = [
    "FinderInfo",
    "_kMDItemUserTags",
    "kMDItemAuthors",
    "kMDItemComment",
    "kMDItemCopyright",
    "kMDItemCreator",
    "kMDItemDescription",
    "kMDItemDownloadedDate",
    "kMDItemDueDate",
    "kMDItemFSIsStationery",
    "kMDItemFinderComment",
    "kMDItemHeadline",
    "kMDItemKeywords",
    "kMDItemParticipants",
    "kMDItemProjects",
    "kMDItemStarRating",
    "kMDItemSubject",
    "kMDItemTitle",
    "kMDItemUserTags",
    "kMDItemVersion",
    "kMDItemWhereFroms",
    "OSXPhotosDetectedText",
]
